#include "ringbuffer_internal.h"





/****************************************************************************/
atom_t ringbuffer_new( void **ringbuffer_state, atom_t number_elements, int (*user_data_init_function)(void **user_data, void *user_state), void *user_state )
{
  struct ringbuffer_state
    *rs;

  volatile struct ringbuffer_element
    *re[RINGBUFFER_PAC_SIZE];

  atom_t
    ringbuffer_count = 0;

  struct ringbuffer_internal_init_state
    riis;

  void
    *freelist_element;

  assert( ringbuffer_state != NULL );
  // number_elements can be any value in its range
  // user_data_init_function can be NULLs

  rs = *ringbuffer_state = abstraction_aligned_malloc( sizeof(struct ringbuffer_state), ALIGN_DOUBLE_POINTER );

  if( rs != NULL )
  {
    riis.user_data_init_function = user_data_init_function;
    riis.user_state = user_state;

    // TRD : the size of the freelist is the size of the ringbuffer (+1 for the leading dummy element in the queue, which is hidden from the caller)
    ringbuffer_count = freelist_new( &rs->freelist_state, number_elements+1, ringbuffer_internal_freelist_init_function, &riis );

    if( rs->freelist_state != NULL )
    {
      // TRD : we may have allocated the freelist but not a single element
      if( ringbuffer_count == 0 )
      {
        freelist_delete( rs->freelist_state, ringbuffer_internal_freelist_delete_function, NULL );
        abstraction_aligned_free( *ringbuffer_state );
        *ringbuffer_state = NULL;
      }

      if( ringbuffer_count >= 1 )
      {
        ringbuffer_count--;

        freelist_pop( rs->freelist_state, &freelist_element );
        freelist_get_user_data_from_element( freelist_element, (void **) &re[RINGBUFFER_POINTER] );

        /* TRD : we can use a counter of 0 here since this is the initial dummy element
                 it cannot be seen on another pointer without having been dequeued and then enqueued
                 and doing so will set its counter
        */
        re[RINGBUFFER_COUNTER] = 0;

        re[RINGBUFFER_POINTER]->next[RINGBUFFER_POINTER] = NULL;
        re[RINGBUFFER_POINTER]->next[RINGBUFFER_COUNTER] = 0;
        re[RINGBUFFER_POINTER]->freelist_element = freelist_element;
        re[RINGBUFFER_POINTER]->user_data = NULL;

        rs->enqueue[RINGBUFFER_POINTER] = rs->dequeue[RINGBUFFER_POINTER] = re[RINGBUFFER_POINTER];
        rs->enqueue[RINGBUFFER_COUNTER] = rs->dequeue[RINGBUFFER_COUNTER] = re[RINGBUFFER_COUNTER];
      }
    }

    if( rs->freelist_state == NULL )
    {
      abstraction_aligned_free( *ringbuffer_state );
      *ringbuffer_state = NULL;
    }
  }

  return( ringbuffer_count );
}





/****************************************************************************/
int ringbuffer_internal_freelist_init_function( void **user_data, void *user_state )
{
  struct ringbuffer_internal_init_state
    *riis;

  struct ringbuffer_element
    *re;

  int
    rv = 0;

  assert( user_data != NULL );
  assert( user_state != NULL );

  riis = (struct ringbuffer_internal_init_state *) user_state;

  // TRD : here we allocate our ringbuffer element into the user_data of the freelist element
  *user_data = re = abstraction_aligned_malloc( sizeof(struct ringbuffer_element), ALIGN_DOUBLE_POINTER );

  if( *user_data != NULL )
  {
    if( riis->user_data_init_function == NULL )
      rv = 1;

    // TRD : here we check to see if the user wishes to init the user data of the ringbuffer element
    if( riis->user_data_init_function != NULL )
      rv = riis->user_data_init_function( &re->user_data, riis->user_state );
  }

  return( rv );
}

