#include "internal.h"





/****************************************************************************/
#if (defined _WIN32 && defined _MSC_VER && !defined NTDDI_VERSION)

  /* TRD : any Windows (user-mode) on any CPU with the Microsoft C compiler

           _WIN32          indicates 64-bit or 32-bit Windows
           _MSC_VER        indicates Microsoft C compiler
           !NTDDI_VERSION  indicates Windows user-mode
  */

  int abstraction_thread_start( thread_state_t *thread_state, thread_function_t thread_function, void *thread_user_state )
  {
    int
      rv = 0;

    DWORD
      thread_id;

    *thread_state = CreateThread( NULL, 0, thread_function, thread_user_state, NO_FLAGS, &thread_id );

    if( *thread_state != NULL )
      rv = 1;

    return( rv );
  }

#endif





/****************************************************************************/
#if (defined _WIN32 && defined _MSC_VER && defined NTDDI_VERSION)

  /* TRD : any Windows on any CPU with the Microsoft C compiler

           _WIN32         indicates 64-bit or 32-bit Windows
           _MSC_VER       indicates Microsoft C compiler
           NTDDI_VERSION  indicates Windows kernel
  */

  int abstraction_thread_start( thread_state_t *thread_state, thread_function_t thread_function, void *thread_user_state )
  {
    int
      rv = 0;

    NTSTATUS
      nts;

    nts = PsCreateSystemThread( thread_state, THREAD_ALL_ACCESS, NULL, NULL, NULL, thread_function, thread_user_state );

    if( nts == STATUS_SUCCESS )
      rv = 1;

    return( rv );
  }

#endif





/****************************************************************************/
#if (defined __unix__)

  /* TRD : any UNIX on any CPU with any compiler

           I assumed pthreads is available on any UNIX.

           __unix__   indicates Solaris, Linux, HPUX, etc
  */

  int abstraction_thread_start( thread_state_t *thread_state, thread_function_t thread_function, void *thread_user_state )
  {
    int
      rv;

    rv = pthread_create( thread_state, NULL, thread_function, thread_user_state );

    return( !rv );
  }

#endif

