#include "internal.h"





/****************************************************************************/
void queue_test( void )
{
  unsigned int
    loop,
    cpu_count;

  void
    *queue_state;

  atom_t
    queue_count;

  thread_state_t
    *thread_handles;

  /* TRD : there are 5 tests

           1. one reader thread per CPU
              - queue always empty
           2. one writer thread per CPU
              - queue always full
           3. one reader and one writer thread per CPU
              - queue balanced
           4. one reader and two writer threads per CPU
              - queue grows
           5. two reader and one writer thread per CPU
              - queue tends to empty
  */

  cpu_count = abstraction_cpu_count();

  printf( "\n"
          "Queue test\n"
          "==========\n" );

  // TRD : 1. one reader thread per CPU

  printf( "\n"
          "1. one reader thread per CPU\n"
          "============================\n" );

  queue_count = queue_new( &queue_state, 10000 );

  if( queue_count != 10000 )
  {
    printf( "queue_new() returned %llu\n", (unsigned long long int) queue_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], queue_internal_thread_reader, queue_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  queue_delete( queue_state, NULL, NULL );

  free( thread_handles );

  // TRD : 2. one writer thread per CPU

  printf( "\n"
          "2. one writer thread per CPU\n"
          "============================\n" );

  queue_count = queue_new( &queue_state, 10000 );

  if( queue_count != 10000 )
  {
    printf( "queue_new() returned %llu\n", (unsigned long long int) queue_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], queue_internal_thread_writer, queue_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  queue_delete( queue_state, NULL, NULL );

  free( thread_handles );

  // TRD : 3. one reader and one writer thread per CPU

  printf( "\n"
          "3. one reader and one writer thread per CPU\n"
          "===========================================\n" );

  queue_count = queue_new( &queue_state, 10000 );

  if( queue_count != 10000 )
  {
    printf( "queue_new() returned %llu\n", (unsigned long long int) queue_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 2 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], queue_internal_thread_reader, queue_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], queue_internal_thread_writer, queue_state );
  }

  for( loop = 0 ; loop < cpu_count * 2 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  queue_delete( queue_state, NULL, NULL );

  free( thread_handles );

  // TRD : 4. one reader and two writer threads per CPU

  printf( "\n"
          "4. one reader and two writer threads per CPU\n"
          "============================================\n" );

  queue_count = queue_new( &queue_state, 10000 );

  if( queue_count != 10000 )
  {
    printf( "queue_new() returned %llu\n", (unsigned long long int) queue_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], queue_internal_thread_reader, queue_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], queue_internal_thread_writer, queue_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], queue_internal_thread_writer, queue_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  queue_delete( queue_state, NULL, NULL );

  free( thread_handles );

  // TRD : 5. two reader and one writer thread per CPU

  printf( "\n"
          "5. two reader and one writer thread per CPU\n"
          "===========================================\n" );

  queue_count = queue_new( &queue_state, 10000 );

  if( queue_count != 10000 )
  {
    printf( "queue_new() returned %llu\n", (unsigned long long int) queue_count );
    return;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], queue_internal_thread_reader, queue_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], queue_internal_thread_reader, queue_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], queue_internal_thread_writer, queue_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  queue_delete( queue_state, NULL, NULL );

  free( thread_handles );

  return;
}





/****************************************************************************/
thread_return_t queue_internal_thread_reader( void *queue_state )
{
  void
    *user_data;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( queue_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    if( queue_dequeue(queue_state, &user_data) )
      count++;
  }

  printf( "read count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t queue_internal_thread_writer( void *queue_state )
{
  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( queue_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    if( queue_enqueue(queue_state, NULL) )
      count++;
  }

  printf( "write count = %llu\n", count );

  return( (thread_return_t) EXIT_SUCCESS );
}

